#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <process.h>
#include <share.h>
#include <time.h>
#include <strstrea.h>

#include "language.h"

#include "misc.h"
#include "process.h"
#include "filetype.h"
#include "execute.h"
#include "rexx.h"
#include "session.h"

SessionItem::SessionItem( ProcessItemType _type, char *_mask, Schedule *_sh ) :
  type( _type ),
  mask( newStr( _mask ) ),
  sh( _sh ),
  ok( 1 )
{
  char *ptr( "" );

  if( type==maskProcessItemType )
  {
    RegularExpresion expr( mask );

    ok = expr.isValid();

    if( !ok )
      ptr = expr.getRegErrorMsg();
  }

  errMsg = newStr( ptr );
}

SessionItem::~SessionItem()
{
  delete mask;
  delete sh;
  delete errMsg;
}

int SessionItem::equal( const BaseObject &t ) const
{
  return type == (*(const SessionItem *)&t).type;
}

void SessionItem::print( ostream& s ) const
{
  s << "Mask:" << ( mask==NULL ? "<NULL>" : mask )
    << ",Type:";

  switch( type )
  {
  case unknownProcessItemType:
    s << "Unknown";
    break;

  case fileProcessItemType:
    s << "File";
    break;

  case filesProcessItemType:
    s << "Files";
    break;

  case maskProcessItemType:
    s << "Mask";
    break;

  case netMailProcessItemType:
    s << "NetMail";
    break;

  case arcMailProcessItemType:
    s << "ArcMail";
    break;

  default:
    s << "<*Who known?*>";
  }

  s << ",Do:" ;

  if( sh!=NULL )
    s << *sh ;
  else
    s << "<NONE>";
}

char *SessionItem::nameOfClass() const
{
  return "SessionItem";
}

char *SessionItem::toString( char *buf, size_t size ) const
{
  memset( buf, 0, size );
  
  ostrstream o( buf, size-1 );
  
  switch( type )
  {
  case unknownProcessItemType:
    o << "Unknown";
    break;

  case fileProcessItemType:
    o << "File " << mask;
    break;

  case maskProcessItemType:
    o << "Mask " << mask;
    break;

  case filesProcessItemType:
    o << "Files";
    break;

  case netMailProcessItemType:
    o << "NetMail";
    break;

  case arcMailProcessItemType:
    o << "ArcMail";
    break;
  }
  
  return buf;
}

int SessionItem::compare( register char *testIni ) const
{
  char *test = fileName( testIni );

  switch( type )
  {
  case fileProcessItemType:
    return !stricmp( mask, test ) ? 1 : 0;

  case maskProcessItemType:
    if( !ok )
      return 0;
    else
    {
      RegularExpresion expr( mask );

      return expr.maskMach( test );
    }
  }

  return 0;
}

ProcessList::ProcessList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

SessionItem* ProcessList::Find( SessionItem& item, LockMode lock ) const
{
  return (SessionItem*)Collection::Find( item, lock );
}
    
SessionItem& ProcessList::operator [] ( const int pos ) const
{
  return *(SessionItem *)&(Collection::operator[]( pos ));
}

ProcessIterator::ProcessIterator( ProcessList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

SessionItem& ProcessIterator::operator () () const
{
  return *(SessionItem *)&(CollectionIterator::operator()());
}

int doMask( SessionItem& cur, ReceivedFileList& in, FIDOAddress &adr,
            long elapsed )
{
  int rc( 0 );

  for( ReceivedFileIterator list( in ); list; list++ )
  {
    char *name( list().getName() );

    if( cur.compare( name ) )
    {
      int rc1( cur.process( adr, elapsed, name ) );

      if( !rc && rc1 )
        rc = rc1;

      if( (miscInfo.state&ProcessOnce)==ProcessOnce )
        break;
    }
  }

  return rc;
}

int doType( SessionItem& cur, ReceivedFileList& in, FileType type,
            FIDOAddress &adr, long elapsed )
{
  int rc( 0 );

  for( ReceivedFileIterator list( in ); list; list++ )
  {
    FileType foundType;
    char *name( list().getName() );

    msgOfFileType( name, &foundType );

    if( foundType==type )
    {
      rc = cur.process( adr, elapsed, name );
      break;
    }
  }

  return rc;
}

void processInbound( ReceivedFileList& in, ProcessList* table,
                     FIDOAddress &adr, long elapsed )
{
  if( table==NULL )
    return;

  if( !in.GetItemsInContainer() )
    return;

  int rc = 0,
      rc1 = 0;

  for( ProcessIterator list( *table ); list; list++ )
  {
    SessionItem& cur = list();

    switch( cur.getType() )
    {
     case fileProcessItemType:
     case maskProcessItemType:
       rc1 = doMask( cur, in, adr, elapsed );
       break;

     case netMailProcessItemType:
       rc1 = doType( cur, in, FileMail, adr, elapsed );
       break;

     case arcMailProcessItemType:
       rc1 = doType( cur, in, FileXMail, adr, elapsed );
       break;

     case filesProcessItemType:
       rc1 = doType( cur, in, FileFile, adr, elapsed );
       break;

     default:
       log( "processInbound:  ProcessItemType" );
    }

    if( rc1 && !rc )
      rc = rc1;
  }


  if( rc>OSReturnCode )
    OSReturnCode = rc;
}

void processImmediate( char *name )
{
  long elapsed = getTimeMark()-sessionStart;
  FileType foundType;
  msgOfFileType( name, &foundType );

  int rc = 0,
      rc1 = 0;

  for( ProcessIterator list( *immediateTable ); list; list++ )
  {
    SessionItem& cur = list();

    switch( cur.getType() )
    {
     case fileProcessItemType:
     case maskProcessItemType:
       if( cur.compare( name ) )
         rc1 = cur.process( *primaryAka, elapsed, name );
       break;

     case netMailProcessItemType:
       if( foundType==FileMail )
         rc1 = cur.process( *primaryAka, elapsed, name );
       break;

     case arcMailProcessItemType:
       if( foundType==FileXMail )
         rc1 = cur.process( *primaryAka, elapsed, name );
       break;

     case filesProcessItemType:
       if( foundType==FileFile )
         rc1 = cur.process( *primaryAka, elapsed, name );
       break;

     default:
       log( "processImmediate:  ProcessItemType" );
    }

    if( rc1 && !rc )
      rc = rc1;
  }

  if( rc>OSReturnCode )
    OSReturnCode = rc;
}
