#include <stdio.h>
#include <stdlib.h>
#include <builtin.h>
#include <process.h>

#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#define INCL_DOSQUEUES
#define INCL_DOSSEMAPHORES
#include <os2.h>

#include "language.h"

#include "misc.h"
#include "queue.h"
#include "nodelist.h"
#include "semaphor.h"

static HMTX resourceSema = 0;
static long sequenceCall = 0;

int ResourceItem::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const ResourceItem *)&t).name );
}

void ResourceItem::print( ostream& s ) const
{
  s << name << ", handle:" << handle;
}

char *ResourceItem::nameOfClass() const
{
  return "ResourceItem";
}

char *ResourceList::findResourceName( unsigned long sema )
{
  for( ResourceIterator list( *this ); list; list++ )
  {
    ResourceItem& cur( list() );

    if( cur.getHandle()==sema )
      return cur.getName();
  }

  return "*unknown*";
}

ResourceItem::ResourceItem( unsigned long _h, char *_n ) :
  handle( _h ),
  name( newStr( _n ) )
{}

ResourceItem::~ResourceItem()
{
  delete name;
}

ResourceList::ResourceList() :
  Collection( 20, 10, SemaphoreModeNoNeed )
{
}

ResourceItem* ResourceList::Find( ResourceItem& item, LockMode lock ) const
{
  return (ResourceItem*)Collection::Find( item, lock );
}
    
ResourceItem& ResourceList::operator [] ( const int pos ) const
{
  return *(ResourceItem*)&(Collection::operator[]( pos ));
}

ResourceIterator::ResourceIterator( ResourceList& ini ) :
  CollectionIterator( *(Collection*)&ini )
{
}

ResourceItem& ResourceIterator::operator () () const
{
  return *(ResourceItem*)&(CollectionIterator::operator()());
}

int initResourceSemaphore()
{
  char sname[ 80 ];
  APIRET rc;

  resourceSema = 0;

  for( rc = ERROR_DUPLICATE_NAME; rc==ERROR_DUPLICATE_NAME;
       rc = DosCreateMutexSem( (PSZ)sname, &resourceSema, DC_SEM_SHARED, 0 ) )
  {
    sprintf( sname, "\\SEM32\\Brake!\\Resource.Semaphore.%08lu", ++sequenceCall );
    capitalize( sname );
  }

  return int( rc );
}

void addResourceItem( ResourceList &list, unsigned long handle,
                      char *name )
{
  APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

  if( rc )
  {
    log( "*BUG* addResourceItem: Unable to request resource list semaphore: %d", rc );
    os2Message( rc );
  }

  list.AddAtTail( *new ResourceItem( handle, name ) );
  rc = DosReleaseMutexSem( resourceSema );

  if( rc )
  {
    log( "*BUG* addResourceItem: Unable to release resource list semaphore: %d", rc );
    os2Message( rc );
  }

}

void detachResourceItem( ResourceList &list, unsigned long handle )
{
  if( !handle || handle==0xffffffffl )
    return;

  ResourceItem *item = NULL;

  APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

  if( rc )
  {
    log( "*BUG* detachResourceItem: Unable to request resource list semaphore: %d", rc );
    os2Message( rc );
  }

  for( ResourceIterator i( list ); i; i++ )
    if( i().getHandle()==handle )
      item = &i();

  if( item==NULL )
    log( "*BUG* detachResourceItem( DoubleList=%08x, ULONG=%lu ) no resource found in list!",
         &list, handle );
  else
    list.Remove( *item );

  rc = DosReleaseMutexSem( resourceSema );

  if( rc )
  {
    log( "*BUG* detachResourceItem: Unable to release resource list semaphore: %d", rc );
    os2Message( rc );
  }
}

unsigned long createMutexSemaphore( char *tag )
{
  char *sname = (char *)_alloca( 80+strlen(tag) );
  unsigned long sema = 0;

  for( APIRET rc( ERROR_DUPLICATE_NAME ); rc==ERROR_DUPLICATE_NAME;
       rc = DosCreateMutexSem( (PSZ)sname, &sema, DC_SEM_SHARED, 0 ) )
  {
    sprintf( sname, "\\SEM32\\Brake!\\%s\\%08lu", tag, ++sequenceCall );
    capitalize( sname );
  }

  if( !rc )
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* createMutexSemaphore: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    mutexSemList.AddAtTail( *new ResourceItem( sema, sname ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* createMutexSemaphore: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }
  }

  return rc ? 0xffffffffl : sema;
}

void deleteMutexSemaphore( unsigned long sema )
{
  if( sema!=0xffffffffl )
  {
    APIRET rc = DosCloseMutexSem( sema );

    if( rc )
    {
      log( "*BUG* deleteMutexSemaphore: Unable to close mutex semaphore #%lu: %d", sema, rc );
      os2Message( rc );
    }

    detachResourceItem( mutexSemList, sema );
  }
}

unsigned long createEventSemaphore( unsigned long *sema, char *tag )
{
  char *sname = (char *)_alloca( 80+strlen(tag) );

  *sema = 0;

  for( APIRET rc( ERROR_DUPLICATE_NAME ); rc==ERROR_DUPLICATE_NAME;
       rc = DosCreateEventSem( (PSZ)sname, sema, DC_SEM_SHARED, 0 ) )
  {
    sprintf( sname, "\\SEM32\\Brake!\\%s\\%08lu", tag, ++sequenceCall );
    capitalize( sname );
  }

  if( !rc )
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* createEventSemaphore: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    eventSemList.AddAtTail( *new ResourceItem( *sema, sname ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* createEventSemaphore: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }
  }

  return rc;
}

void deleteEventSemaphore( unsigned long sema )
{
  if( sema!=0xffffffffl )
  {
    APIRET rc = DosCloseEventSem( sema );

    if( rc )
    {
      log( "*BUG* deleteEventSemaphore: Unable to close event semaphore %s: %d",
           eventSemList.findResourceName( sema ), rc );
      os2Message( rc );
    }

    detachResourceItem( eventSemList, sema );
  }
}

unsigned long createQueue( unsigned long *sema, char *tag, char **qn )
{
  char *sname = (char *)_alloca( 80+strlen(tag) );

  *sema = 0;

  for( APIRET rc( ERROR_QUE_DUPLICATE ); rc==ERROR_QUE_DUPLICATE;
       rc = DosCreateQueue( sema, QUE_FIFO|QUE_CONVERT_ADDRESS, (PSZ)sname ) )
  {
    sprintf( sname, "\\QUEUES\\Brake!\\%s\\%08lu", tag, ++sequenceCall );
    capitalize( sname );
  }

  if( !rc )
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* createQueue: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    queueList.AddAtTail( *new ResourceItem( *sema, sname ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* createQueue: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }

    if( qn )
      *qn = newStr( sname );
  }
  else if( qn )
    *qn = NULL;

  return rc;
}

void deleteQueue( unsigned long *queue )
{
  APIRET rc = DosCloseQueue( *queue );

  if( rc )
  {
    log( "*BUG* deleteQueue: Unable to close queue %a: %d",
         queueList.findResourceName( *queue ), rc );
    os2Message( rc );
  }

  detachResourceItem( queueList, *queue );
  *queue = 0xffffffffl;
}

int checkUsage( char *name )
{
  char pref[] = "\\SEM32\\Brake!",
       *sname = (char *)_alloca( strlen(name)+sizeof(pref)+1 );

  strcpy( sname, pref );

  char drive[_MAX_DRIVE],
       dir  [_MAX_DIR],
       fname[_MAX_FNAME],
       ext  [_MAX_EXT];

  _splitpath( name, drive, dir, fname, ext );

  if( *dir!='\\' )
    strcat( sname, "\\" );

  strcat( sname, dir );
  strcat( sname, fname );
  strcat( sname, ext );
  capitalize( sname );

  HEV sema = 0;

  APIRET rc = DosCreateMutexSem( (PSZ)sname, &sema, DC_SEM_SHARED, 0 );

  if( rc && rc!=ERROR_DUPLICATE_NAME )
  {
    log( Msg_ErrUnableCreSem, sname, rc );
    os2Message( rc );
  }

  if( !rc )
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* checkUsage: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    mutexSemList.AddAtTail( *new ResourceItem( sema, sname ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* checkUsage: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }
  }

  return rc == ERROR_DUPLICATE_NAME;
}

char *skipPath( register char *name, register char *pre );

void brakeSemaphoreCreate( char *name )
{
  HEV sema = 0;
  APIRET rc = DosCreateEventSem( (PSZ)name, &sema, DC_SEM_SHARED, 0 );

  if( rc )
  {
    log( Msg_SemNotCreated, name, rc );
    os2Message( rc );
  }
  else
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* brakeSemaphoreCreate: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    eventSemList.AddAtTail( *new ResourceItem( sema, name ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* brakeSemaphoreCreate: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }

    log( Msg_SemCreated, skipPath( name, miscInfo.semaphore ) );
  }
}

void brakeSemaphoreRemove( char *name )
{
  ResourceItem test( 0, name ),
               *found( eventSemList.Find( test ) );

  if( found==NULL )
  {
    log( Msg_SemDelNotExist, name );
    return;
  }

  APIRET rc = DosCloseEventSem( found->getHandle() );

  if( rc )
  {
    log( Msg_SemNotRemoved, name, rc );
    os2Message( rc );
  }
  else
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* brakeSemaphoreRemove: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    eventSemList.Remove( *found );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* brakeSemaphoreRemove: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }

    log( Msg_SemRemoved, skipPath( name, miscInfo.semaphore ) );
  }
}

void brakeSemaphorePost( char *name )
{
  ResourceItem test( 0, name ),
               *found( eventSemList.Find( test ) );
  HEV hand(0);
  APIRET rc;

  if( found==NULL )
  {
    rc = DosOpenEventSem( name, &hand );

    if( rc )
    {
      log( Msg_SemPostNotExist, name, rc );
      os2Message( rc );
      return;
    }
  }
  else
    hand = found->getHandle();

  rc = DosPostEventSem( hand );

  if( rc && rc!=ERROR_ALREADY_POSTED )
  {
    log( Msg_SemNotPosted, name, rc );
    os2Message( rc );
  }
  else
    log( Msg_SemPosted, skipPath( name, miscInfo.semaphore ) );

  if( found==NULL )
    DosCloseEventSem( hand );
}

void brakeSemaphoreReset( char *name )
{
  ResourceItem test( 0, name ),
               *found( eventSemList.Find( test ) );
  HEV hand(0);
  APIRET rc;

  if( found==NULL )
  {
    rc = DosOpenEventSem( name, &hand );

    if( rc )
    {
      log( Msg_SemResetNotExist, name, rc );
      os2Message( rc );
      return;
    }
  }
  else
    hand = found->getHandle();

  ULONG count;

  rc = DosResetEventSem( hand, &count );

  if( rc && rc!=ERROR_ALREADY_RESET )
  {
    log( Msg_SemNotReseted, name, rc );
    os2Message( rc );
  }
  else
    log( Msg_SemReseted, skipPath( name, miscInfo.semaphore ), count );

  if( found==NULL )
    DosCloseEventSem( hand );
}

void brakeSemaphoreWait( char *name )
{
  ResourceItem test( 0, name ),
               *found( eventSemList.Find( test ) );
  HEV hand(0);
  APIRET rc;

  if( found==NULL )
  {
    rc = DosOpenEventSem( name, &hand );

    if( rc )
    {
      log( Msg_SemWaitNotExist, name, rc );
      os2Message( rc );
      return;
    }
  }
  else
    hand = found->getHandle();

  ULONG count=0;

  rc = DosQueryEventSem( hand, &count );

  if( rc )
  {
    log( Msg_SemWaitErr, name, rc );
    os2Message( rc );
  }
  else
    if( !count )
      log( Msg_SemWait, skipPath( name, miscInfo.semaphore ) );

  rc = DosWaitEventSem( hand, (ULONG)SEM_INDEFINITE_WAIT );

  if( rc )
  {
    log( Msg_SemNotWaited, skipPath( name, miscInfo.semaphore ), rc );
    os2Message( rc );
  }
  else
    log( Msg_SemWaited+(count ? 1 : 0), skipPath( name, miscInfo.semaphore ) );

  if( found==NULL )
    DosCloseEventSem( hand );
}

int busySemaphoreCreate( char *name )
{
  HEV sema = 0;
  APIRET rc = DosCreateMutexSem( (PSZ)name, &sema, DC_SEM_SHARED, 0 );

  if( rc )
  {
    log( Msg_BusySemNotCreated, name, rc );
    os2Message( rc );
  }
  else
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* busySemaphoreCreate: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    mutexSemList.AddAtTail( *new ResourceItem( sema, name ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* busySemaphoreCreate: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }
  }

  return rc==0;
}

int busySemaphoreRemove( char *name )
{
  ResourceItem test( 0, name ),
               *found( mutexSemList.Find( test ) );

  if( found==NULL )
  {
    log( Msg_BusySemDelNotExist, name );
    return 0;
  }

  APIRET rc = DosCloseMutexSem( found->getHandle() );

  if( rc )
  {
    log( Msg_BusySemNotRemoved, name, rc );
    os2Message( rc );
  }
  else
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* busySemaphoreRemove: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    mutexSemList.Remove( *found );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* busySemaphoreRemove: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }
  }

  return rc==0;
}

int busySemaphoreExist( char *name )
{
  ResourceItem test( 0, name ),
               *found( mutexSemList.Find( test ) );

  if( found!=NULL )
    return 1;

  HEV hand(0);
  APIRET rc = DosOpenMutexSem( name, &hand );

  if( !rc )
    DosCloseMutexSem( hand );

  return rc==0;
}

int createNamedMutex( char *name, unsigned long *sema )
{
  *sema = 0;

  APIRET rc = DosOpenMutexSem( name, sema );

  if( rc )
    rc = DosCreateMutexSem( name, sema, DC_SEM_SHARED, 0 );

  if( rc )
  {
    *sema = 0xffffffffl;
    log( Msg_ErrUnableCreSem, name, rc );
    os2Message( rc );
  }
  else
  {
    APIRET rc = DosRequestMutexSem( resourceSema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* createNamedMutex: Unable to request resource list semaphore: %d", rc );
      os2Message( rc );
    }

    mutexSemList.AddAtTail( *new ResourceItem( *sema, name ) );
    rc = DosReleaseMutexSem( resourceSema );

    if( rc )
    {
      log( "*BUG* createNamedMutex: Unable to release resource list semaphore: %d", rc );
      os2Message( rc );
    }
  }

  return int( rc );
}

char *skipPath( register char *name, register char *pre )
{
  if( begin( name, pre ) )
    name += strlen( pre );

  return name;
}

void lockMutexSemaphore( unsigned long sema )
{
  if( sema!=0xffffffffl )
  {
    APIRET rc = DosRequestMutexSem( sema, (ULONG)SEM_INDEFINITE_WAIT );

    if( rc )
    {
      log( "*BUG* lockMutexSemaphore: Unable to request mutex semaphore %s: %d",
           findMutexSemaphoreName( sema ), rc );
      os2Message( rc );
    }
  }
}

void unlockMutexSemaphore( unsigned long sema )
{
  if( sema!=0xffffffffl )
  {
    APIRET rc = DosReleaseMutexSem( sema );

    if( rc && rc!=ERROR_NOT_OWNER )
    {
      log( "*BUG* unlockMutexSemaphore: Unable to release mutex semaphore %s: %d",
           findMutexSemaphoreName( sema ), rc );
      os2Message( rc );
    }
  }
}

char *findMutexSemaphoreName( unsigned long sema )
{
  return mutexSemList.findResourceName( sema );
}

int queryMutexSemaphore( unsigned long sema )
{
  if( sema==0xffffffffl )
    return 0;

  PID pid;
  TID tid;
  ULONG count;
  APIRET rc( DosQueryMutexSem( sema, &pid, &tid, &count ) );

  if( rc )
  {
    log( Msg_SemWaitErr, findMutexSemaphoreName( sema ), rc );
    os2Message( rc );
    return 0;
  }

  if( !count )
    return 0;

  return (pid==getpid() && tid==getMyTid()) ? 0 : 1;
}