//----------------------------------------------------------------------------
//
// Copyright (c) 1995 by Bad Data, INC.
// All rights reserved.
//
// Written by John A. Gladkih.
//
// Class window - implementation.
//

#include <string.h>
#include <stdio.h>
#include <builtin.h>

#include "window.h"
#include "semaphor.h"
#include "misc.h"

#include "language.h"

Color::Color( const Color& col )
{
  fore = col.fore;
  back = col.back;
}

Color::Color( int _fore, int _back )
{
  fore = (unsigned char)(_fore&0xf);
  back = (unsigned char)(_back&0x7);
}

Color::Color( unsigned char cl )
{
  back = (unsigned char)((cl>>4)&0xf);
  fore = (unsigned char)(cl&0xf);
}

Color::Color( unsigned int cl )
{
  back = (unsigned char)((cl>>4)&0xf);
  fore = (unsigned char)(cl&0xf);
}

Color::Color( int cl )
{
  back = (unsigned char)((cl>>4)&0xf);
  fore = (unsigned char)(cl&0xf);
}

Color& Color::operator=( int _fore )
{
  fore = (unsigned char)_fore;
  return *this;
}

Color& Color::operator=( const Color& col )
{
  fore=col.fore;
  back=col.back;

  return *this;
}

Color::operator unsigned char() const
{
  return (unsigned char)(fore|(back<<4));
}

Color::operator int() const
{
  return int(operator unsigned char())&0xff;
}

unsigned long Window::mutex = 0xffffffffl;
int Window::needLock = 1;

void Window::lock()
{
  if( !detachedSession && needLock && mutex!=0xffffffffl && mutex!=0xfffffffel )
    lockMutexSemaphore( mutex );
}

void Window::unlock()
{
  if( !detachedSession && needLock && mutex!=0xffffffffl && mutex!=0xfffffffel )
    unlockMutexSemaphore( mutex );
}

Window::Window( int _left, int _top,
                       int _width, int _height,
                       const Color& _col,
		       char* _title,
                       int _brd,
                       int _shd, const Color _dropClr ) :
  color( _col ),
  saveX( 0 ),
  saveY( 0 ),
  physX( -1 ),
  physY( -1 ),
  visible( 0 ),
  delta( _brd==BS_NOBORDER ? 0 : 1 ),
  width( _width ),
  height( _height ),
  left( _left ),
  top( _top ),
  saveWnd( NULL ),
  scrolled( 1 ),
  wnd( 0 ),
  pWin( 0 )
{
  top += delta;
  left += delta;
  width -= delta<<1;
  height -= delta<<1;

  if( !detachedSession )
  {
    if( mutex==0xffffffffl )
    {
      mutex = createMutexSemaphore( "Window" );

      if( mutex==0xffffffffl )
      {
        log( Msg_ErrWindowMutex );
        mutex = 0xfffffffel;
      }
    }

    wnd = new WINDOW;
    lock();
    WINDOW *pp = Z_CreateWindow(  wnd,
				_height, _width,
				_left-1, _top-1,
				(int(_col)<<8)|0x20,
				_brd, _col.fore, _col.back,
				_shd, _dropClr,
				_title, NULL );
    pWin = Z_DefineWindow( pp );
    unlock();
  }
}

Window::~Window()
{
  if( !detachedSession )
  {
    lock();
    Z_CloseWindow( pWin );
    unlock();
    delete wnd;
  }

  delete saveWnd;
}

void Window::setPhysicalCursor( int force )
{
  if( !detachedSession )
    if( saveX!=physX || saveY!=physY || force )
    {
      lock();
      physX = saveX;
      physY = saveY;
      Z_SetCursor( pWin, saveX, saveY );
      unlock();
    }
}

void Window::gotoXY( int x, int y )
{
  if( !detachedSession )
  {
    --x;
    --y;

    if( x>=0 && x<getWidth() )
      saveX = x;

    if( y>=0 && y<getHeight() )
      saveY = y;
  }
}

void Window::clearLineNoLock()
{
  if( detachedSession )
    return;

  char *w = (char *)_alloca( width+10 );
  int sc( scrolled ),
      xx( saveX ),
      yy( saveY );

  for( int i( saveX ), j(0); i<getWidth(); i++ )
    w[ j++ ]=0x20;

  w[ j ] = 0;

  scrolled=0;
  putStringNoLock( w );

  scrolled = char(sc);
  saveX = xx;
  saveY = yy;
}

Window& Window::move( int x, int y )
{
  if( detachedSession )
    return *this;

  char vis( visible );

  -*this;

  lock();
  left = pWin->XOrg = x-1;
  top = pWin->YOrg = y-1;
  unlock();

  if( vis )
    +*this;

  return *this;
}

Window& Window::centre( int row, char *str )
{
  if( detachedSession )
    return *this;

  int len( strlen( str ) );

  if( len>=width )
    len = 1;
  else
    len = (width-strlen(str))/2+1;

  return (*this)( len, row ) << str;
}

void Window::putChar( char ch, int lockMode )
{
  if( detachedSession )
    return;

  if( lockMode )
    lock();

  switch( ch )
  {
  case '\r':
    saveX = 0;
    break;

  case '\b':
    if( --saveX<0 )
      saveX = 0;
    break;

  default:
    Z_CharOut( pWin, ch, saveX+delta, saveY+delta, color.fore, color.back );

    if( ++saveX>=width )
    {
      if( scrolled )
	saveX = 0;
      else
	saveX--;

  case '\n':

      saveY++;

      if( saveY>=height )
      {
	saveY--;
	scrollUp();
      }
    }
  }

  if( lockMode )
    unlock();
}

void Window::save()
{
  if( detachedSession )
    return;

  register int size = width*height;

  if( saveWnd==NULL )
    saveWnd = new int [ size ];

  lock();
  memcpy( saveWnd, pWin->pVS, size*sizeof(int) );
  unlock();
}

void Window::restore()
{
  if( detachedSession )
    return;

  if( saveWnd==NULL )
    return;

  lock();
  memcpy( pWin->pVS, saveWnd, width*height*sizeof(int) );
  Z_UpdatePhysicalScreen( pWin, 0, 0, width, height );
  unlock();
}

void Window::scrollUp()
{
  if( detachedSession )
    return;

  if( scrolled )
  {
    int *p=pWin->pVS;

    memcpy( p, p+width, width*(height-1)*sizeof(int) );

    for( p += width*(height-1); p < pWin->pVS+width*height; p++ )
      *p = (int(color)<<8)|0x20;

    Z_UpdatePhysicalScreen( pWin, 0, 0, width, height );
  }
}

void Window::putStringNoLock( const char * const stri )
{
  if( detachedSession )
    return;

/*
  for( ; *str; str++ )
  {
    putChar( *str, 0 );

    if( !scrolled )
      if( saveX+1>=width )
        saveX = 0;
  }
*/
  const char * str( stri );
  char *n = (char *)_alloca( strlen(str)+10 );

  while( *str )
  {
    strcpy( n, str );

    int rest( getWidth()-saveX ),
	len( strlen(n) );

    if( len>rest )
      len = rest;

    str += len;

    int cr( ( n[len-1]=='\r' || n[len-2]=='\r' ) ? 1 : 0 );
    int nl( ( n[len-1]=='\n' || n[len-2]=='\n' ) ? 1 : 0 );

    n[ len-cr-nl ] = 0;

    Z_TextOut( pWin, n, saveX+delta, saveY+delta, color.fore, color.back );
    saveX += len-cr-nl;

    if( cr )
      putChar( '\r', 0 );

    if( nl )
      putChar( '\n', 0 );

    if( saveX>=width )
    {
      saveX = 0;
      saveY++;

      if( saveY>=height )
      {
	saveY--;
	scrollUp();
      }
    }
  }
}

void Window::drawTransact( int x, int y, const Color &cl, unsigned long num, int clear )
{
  if( detachedSession )
    return;

  char buf[32];

  sprintf( buf, "%lu", num );
  drawTransact( x, y, cl, buf, clear );
}

void Window::drawTransact( int x, int y, const Color &cl, const char * const str, int clear )
{
  if( detachedSession )
    return;

  lock();

  gotoXY( x, y );
  color = cl;
  putStringNoLock( str );

  if( clear )
    clearLineNoLock();

  unlock();
}

void Window::drawTransactField( int x, int y, const Color &cl,
                                unsigned long num, int clear )
{
  if( detachedSession )
    return;

  char buf[32];

  sprintf( buf, "%lu", num );
  drawTransactField( x, y, cl, buf, clear );
}

void Window::drawTransactField( int x, int y, const Color &cl,
                                const char * const str, int clear )
{
  if( detachedSession )
    return;

  lock();

  gotoXY( x, y );
  color = cl;

  if( clear && putFieldNoLock( str ) )
    clearLineNoLock();

  unlock();
}

Window& Window::operator+()
{
  if( !detachedSession )
    if( !visible )
    {
      lock();
      Z_ShowWindow( pWin );
      visible=1;
      unlock();
    }

  return *this;
}

Window& Window::operator-()
{
  if( !detachedSession )
    if( visible )
    {
      lock();
      Z_HideWindow( pWin );
      visible=0;
      unlock();
    }

  return *this;
}

Window& Window::operator--()
{
  if( !detachedSession )
  {
    lock();
    Z_ClearWindow( pWin );
    saveX=0;
    saveY=0;
    unlock();
  }

  return *this;
}

Window& Window::operator<<( long c )
{
  if( !detachedSession )
  {
    char buffer [32];

    sprintf( buffer, "%ld", c );
    putString( buffer );
  }

  return *this;
}

Window& Window::operator<<( unsigned long c )
{
  if( !detachedSession )
  {
    char buffer [32];

    sprintf( buffer, "%lu", c );
    putString( buffer );
  }

  return *this;
}

Window& Window::operator<<( int c )
{
  if( !detachedSession )
  {
    char buffer [32];

    sprintf( buffer, "%d", c );
    putString( buffer );
  }

  return *this;
}

Window& Window::operator<<( unsigned int c )
{
  if( !detachedSession )
  {
    char buffer [32];

    sprintf( buffer, "%u", c );
    putString( buffer );
  }

  return *this;
}

Window& Window::operator<<( void* c )
{
  if( !detachedSession )
  {
    char buffer [32];

    sprintf( buffer, "%p", c );
    putString( buffer );
  }

  return *this;
}

Window& Window::field( const char * const str )
{
  if( !detachedSession )
  {
    lock();
    putFieldNoLock( (char *)str );
    unlock();
  }

  return *this;
}

int Window::putFieldNoLock( const char * const stri )
{
  if( detachedSession )
    return 0;

  char *str( tmpStr( stri ) );
  int left( getWidth()-getCurrentX()+1 );
  int rc( 1 );

  if( left<=0 )
    return 0;

  if( strlen( str )>left )
  {
    str[ left ] = 0;
    str[ left-1 ] = 26;
    rc = 0;
  }

  putStringNoLock( str );
  return rc;
}

void Window::pop()
{
  if( !detachedSession )
  {
    lock();
    Z_PopWindow( pWin );
    unlock();
  }
}

void Window::push()
{
  if( !detachedSession )
  {
    lock();
    Z_PushWindow( pWin );
    unlock();
  }
}

void Window::putString( const char * const str )
{
  if( detachedSession )
    return;

  lock();
  putStringNoLock( str );
  unlock();
}

void Window::setLockMode( int lock )
{
  needLock=lock;
}

Window& Window::operator<<( signed char c )
{
  if( !detachedSession )
    putChar( char(c) );

  return *this;
}

Window& Window::operator<<( char c )
{
  if( !detachedSession )
    putChar( char(c) );

  return *this;
}

Window& Window::operator<<( const signed char * const  c )
{
  if( !detachedSession )
    putString( (const char * const)c );

  return *this;
}

Window& Window::operator<<( const char * const  c )
{
  if( !detachedSession )
    putString( (const char * const)c );

  return *this;
}

Window& Window::operator()( int x, int y ) 
{
  if( !detachedSession )
  {
    lock();
    gotoXY( x, y );
    unlock();
  }

  return *this;
}

Window& Window::operator()( int y )
{
  if( !detachedSession )
  {
    lock();
    gotoXY( 1, y );
    unlock();
  }

  return *this;
}

void Window::drawTransact( int x, int y, const Color &cl, long num, int clear )
{
  if( !detachedSession )
    drawTransact( x, y, cl, (long unsigned)num, clear );
}

void Window::drawTransact( int x, int y, const Color &cl, int num, int clear )
{
  if( !detachedSession )
    drawTransact( x, y, cl, (long unsigned)num, clear );
}

void Window::drawTransactField( int x, int y, const Color &cl,
                                long num, int clear )
{
  if( !detachedSession )
    drawTransactField( x, y, cl, (long unsigned)num, clear );
}

void Window::drawTransactField( int x, int y, const Color &cl,
                                int num, int clear )
{
  if( !detachedSession )
    drawTransactField( x, y, cl, (long unsigned)num, clear );
}

Window& Window::clearLine()
{
  if( !detachedSession )
  {
    lock();
    clearLineNoLock();
    unlock();
  }

  return *this;
}

void Window::setScroll( char s )
{
  scrolled = s;
}
